from distutils.core import setup

setup(
    name='swampy',
    version='2.1.5',
    author='Allen B. Downey',
    author_email='downey@allendowney.com',
    packages=['swampy'],
    scripts=[],
    url='http://pypi.python.org/pypi/swampy',
    license='LICENSE.txt',
    description='Companion code for Think Python.',
    long_description=open('README.txt').read(),
)
